if select(6, GetAddOnInfo("PitBull4_" .. (debugstack():match("[o%.][d%.][u%.]les\\(.-)\\") or ""))) ~= "MISSING" then return end

local PitBull4 = _G.PitBull4
if not PitBull4 then
error("PitBull4_VisualHelpers requires PitBull4")
end

-- :::::::::::: LOCALS ::::::::::::: -------------------------------------------------------------------------

local BASE_TEXTURE_PATH = [[Interface\AddOns\]] .. debugstack():match("[d%.][d%.][O%.]ns\\(.-)\\[A-Za-z]-%.lua") .. [[\]]

local L = PitBull4.L

local GetLayoutDB = PitBull4.Options.GetLayoutDB

local PitBull4_VisualHelpers = PitBull4:NewModule("VisualHelpers", "AceEvent-3.0", "AceTimer-3.0")
PitBull4_VisualHelpers:SetModuleType("custom")
PitBull4_VisualHelpers:SetName(L["Visual helpers"])
PitBull4_VisualHelpers:SetDescription(L["Show visual helpers for aura groups affecting a unit."])
PitBull4_VisualHelpers:SetDefaults({})

local AceGUI = LibStub("AceGUI-3.0")

PitBull4_VisualHelpers.VH_INDICATORS = {}
local updated_guids = {}
local auras = {}

-- timer for OnUpdate execution control
local timerFrame = PitBull4.Controls.MakeFrame()
timerFrame:Hide()
local timer = 0
timerFrame:SetScript("OnUpdate",function(self, elapsed)
	timer = timer + elapsed
	if timer >= 0.2 then
		PitBull4_VisualHelpers:OnUpdate()
		timer = 0
	end
end)
local timer = 0

local helpers = PitBull4_VisualHelpers.db.profile.global.helpers

-- Fills an array of arrays with the information about the auras
local function reset_auras()
	for a, _ in pairs(auras) do
		auras[a].present = false
		auras[a].data = {}
	end
end
local function update_aura_list(unit)
	reset_auras()
	if not unit then return end
	local index = 1
	local filter = 'HELPFUL'

	local aura, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId, canApplyAura, isBossDebuff
	
	-- Loop through the auras
	while true do
		aura, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId, canApplyAura, isBossDebuff = UnitAura(unit, index, filter)
		if not aura then
			if filter == 'HELPFUL' then
				filter = 'HARMFUL'
				index = 0
			else
				break
			end
		else
			if auras[aura] and (not auras[aura].selfonly or (unitCaster and UnitIsUnit(unitCaster, 'player'))) then
				local dontoverwrite = false
				if auras[aura].present then
					--If multiple of the same aura, show longest or most stacks
					if (expirationTime < auras[aura].data.expirationTime) or (count < auras[aura].data.count) then
						dontoverwrite = true
					end
				end
				if not dontoverwrite then
					auras[aura].present = true
					auras[aura].data = {
						aura = aura,
						icon = icon,
						count = count,
						debuffType = debuffType,
						duration = duration or 0,
						expirationTime = expirationTime or 0,
						unitCaster = unitCaster,
						isStealable = isStealable,
					}
				end
			end
		end
		index = index + 1
	end
end

local function update_frame(frame)
	--print("update_frame "..frame.unit)
	update_aura_list(frame.unit)
	
	local visible
	local bestaura
			
	for helper_name, helper in pairs(helpers) do
		local indicator = PitBull4_VisualHelpers.VH_INDICATORS[helper_name]
		local db = indicator:GetLayoutDB(frame)
			
		if db.enabled then
		
			visible = false
			bestaura = false
			
			for aura, auradata in pairs(helper.auras) do
								
				if auras[aura].present then 
									
					local a = auras[aura]
					-- Zone and caster filter
					if ((not a.zone or a.zone == '') or a.zone == GetRealZoneText()) and ((not a.selfonly) or UnitIsUnit(a.data.unitCaster, 'player')) then
								
						visible = true
						
						if bestaura then
							bestaura.priority = bestaura.priority or 0
							bestaura.duration = bestaura.duration or 0
							bestaura.expirationTime = bestaura.expirationTime or 0
						end
						
						a.priority = a.priority or 0
						a.duration = a.duration or 0
						a.expirationTime = a.expirationTime or 0
						
						if (not bestaura) or
							(a.priority>bestaura.priority) or
							(a.priority==bestaura.priority and (bestaura.duration==0 or bestaura.expirationTime==0 or a.expirationTime>bestaura.expirationTime) and a.duration>0) then
								
							bestaura = a.data
							bestaura.priority = a.priority or 0
						end
						
					end
					
				end
			end
			if (frame[indicator.full_name..'_show'] and not visible) or visible then
				frame[indicator.full_name..'_show'] = visible
				frame[indicator.full_name..'_aura'] = bestaura
				frame[indicator.full_name..'_guid'] = UnitGUID(frame.unit)
				indicator:Update(frame)
			end
		end
		
		indicator = nil
		db = nil
	end
end

-- :::::::::::: PitBull4_VisualHelpers ::::::::::::: -------------------------------------------------------------------------
function PitBull4_VisualHelpers:OnEnable()
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
end

function PitBull4_VisualHelpers:PLAYER_ENTERING_WORLD()
	if not (PitBull4_VisualHelpers.db.profile.global.helpers 
		and next(PitBull4_VisualHelpers.db.profile.global.helpers)) then
		PitBull4_VisualHelpers:CreateDefaultGlobalHelpers()
		helpers = PitBull4_VisualHelpers.db.profile.global.helpers
	end
	self:UpdateIndicators()
	timerFrame:Show()
end

function PitBull4_VisualHelpers:UNIT_AURA(event, unit)
	updated_guids[UnitGUID(unit)] = true
end

function PitBull4_VisualHelpers:RAID_ROSTER_UPDATE()
	for frame in PitBull4:IterateFrames() do
		update_frame(frame)
	end
end
function PitBull4_VisualHelpers:PARTY_MEMBERS_CHANGED()
	for frame in PitBull4:IterateFrames() do
		update_frame(frame)
	end
end

function PitBull4_VisualHelpers:UpdateFrame(frame)
	if #self.VH_INDICATORS==0 then
		PitBull4_VisualHelpers:UpdateIndicators()
	end
	update_frame(frame)
end
function PitBull4_VisualHelpers:ForceUpdate(frame)
	update_frame(frame)
end

function PitBull4_VisualHelpers:OnUpdate()
	if next(updated_guids) then
		for frame in PitBull4:IterateFrames() do
			if updated_guids[frame.guid] then
				update_frame(frame)
			end
		end
	end
	wipe(updated_guids)
end

function PitBull4_VisualHelpers:OnDisable()
	timerFrame:Hide()
	wipe(self.VH_INDICATORS)
end

function PitBull4_VisualHelpers:UpdateIndicators()
	auras = {}
	if not helpers then return end
	for n, h in pairs(helpers) do
		if not PitBull4.modules["VisualHelpers: "..n] then
			self.VH_INDICATORS[n] = self:CreateVisualHelperIndicator(n)
		end
		for a, adata in pairs(h.auras) do
			if not auras[a] then
				auras[a] = {}
				auras[a].zone = adata.zone or ''
				auras[a].priority = adata.priority or 0
				auras[a].selfonly = adata.selfonly or false
				auras[a].present = nil
				auras[a].data = {}
			end
		end
	end
	for frame in PitBull4:IterateFrames() do
		update_frame(frame)
	end
end
